import argparse
import textwrap

from decli import cli


def main_example() -> argparse.ArgumentParser:
    """https://docs.python.org/3/library/argparse.html#example"""
    data = {
        "description": "Process some integers.",
        "arguments": [
            {
                "name": "integers",
                "metavar": "N",
                "type": int,
                "nargs": "+",
                "help": "an integer for the accumulator",
            },
            {
                "name": "--sum",
                "dest": "accumulate",
                "action": "store_const",
                "const": sum,
                "default": max,
                "help": "sum the integers (default: find the max)",
            },
        ],
    }
    parser = cli(data)
    return parser


def complete_example() -> argparse.ArgumentParser:
    data = {
        "prog": "cz",
        "formatter_class": argparse.RawDescriptionHelpFormatter,
        "description": "The software does this and that",
        "epilog": "This is the epilooogpoe  ",
        "arguments": [
            {
                "name": "--debug",
                "action": "store_true",
                "default": False,
                "help": "use debug mode",
            },
            {
                "name": ["-v", "--version"],
                "action": "store_true",
                "default": False,
                "help": "get the installed version",
                "group": "ops",
            },
            {"name": "--save", "group": "ops"},
        ],
        "subcommands": {
            "title": "main",
            "description": "main commands",
            "commands": [
                {
                    "name": "all",
                    "help": "check every values is true",
                    "func": all,
                },
                {
                    "name": ["sum", "s"],
                    "help": "new project",
                    "func": sum,
                    "arguments": [
                        {
                            "name": "integers",
                            "metavar": "N",
                            "type": int,
                            "nargs": "+",
                            "help": "an integer for the accumulator",
                        },
                        {"name": "--name", "nargs": "?"},
                    ],
                },
            ],
        },
    }
    parser = cli(data)
    return parser


def name_or_flags() -> dict:
    """https://docs.python.org/3/library/argparse.html#name-or-flags"""
    data = {
        "prog": "sti",
        "arguments": [{"name": ["-f", "--foo"]}, {"name": "bar"}],
    }
    return data


def compose_clis_using_parents() -> list[argparse.ArgumentParser]:
    """
    Sometimes, several cli share a common set of arguments.
    Rather than repeating the definitions of these arguments,
    one or more parent clis with all the shared arguments can be passed
    to parents= argument to cli.

    https://docs.python.org/3/library/argparse.html#parents
    """
    parent_foo_data = {
        "add_help": False,
        "arguments": [{"name": "--foo-parent", "type": int}],
    }
    parent_bar_data = {
        "add_help": False,
        "arguments": [{"name": "--bar-parent", "type": int}],
    }
    parent_foo_cli = cli(parent_foo_data)
    parent_bar_cli = cli(parent_bar_data)

    parents = [parent_foo_cli, parent_bar_cli]
    return parents


def using_formatter_class() -> dict:
    """https://docs.python.org/3/library/argparse.html#formatter-class"""
    data = {
        "prog": "PROG",
        "formatter_class": argparse.RawDescriptionHelpFormatter,
        "description": textwrap.dedent(
            """\
            Please do not mess up this text!
            --------------------------------
                I have indented it
                exactly the way
                I want it
            """
        ),
    }
    return data


def prefix_chars() -> dict:
    data = {
        "prog": "PROG",
        "prefix_chars": "+",
        "arguments": [{"name": ["+f", "++foo"]}, {"name": "++bar"}],
    }
    return data


def grouping_arguments() -> dict:
    data = {
        "prog": "mycli",
        "arguments": [
            {
                "name": "--new",
                "help": "This does not belong to a group but its a long help",
            },
            {
                "name": "--init",
                "help": "This does not belong to a group but its a long help",
            },
            {
                "name": "--run",
                "group": "app",
                "help": "This does not belong to a group",
            },
            {
                "name": "--build",
                "group": "app",
                "help": "This does not belong to a group",
            },
            {
                "name": ["--install", "--add"],
                "group": "package",
                "metavar": "package_name",
                "nargs": "+",
                "help": "This belongs to a group",
            },
            {
                "name": "--remove",
                "group": "package",
                "help": "This belongs to a group",
            },
            {
                "name": "--why",
                "group": "package",
                "help": "This belongs to a group",
            },
        ],
    }
    return data


def exclusive_group() -> dict:
    data = {
        "prog": "app",
        "arguments": [
            {"name": "--install", "exclusive_group": "ops"},
            {"name": "--purge", "exclusive_group": "ops"},
        ],
    }
    return data



"""Most of argparse examples rebuilt with climp."""

import argparse
import unittest

import pytest

from decli import cli

from . import examples


class Test(unittest.TestCase):
    def test_main_example_ok(self) -> None:
        parser = examples.main_example()
        args = parser.parse_args("1 2 3 4".split())

        assert args.accumulate(args.integers) == 4

    def test_main_example_sums_ok(self) -> None:
        parser = examples.main_example()
        args = parser.parse_args("1 2 3 4 --sum".split())

        assert args.accumulate(args.integers) == 10

    def test_main_example_fails(self) -> None:
        parser = examples.main_example()

        with pytest.raises(SystemExit):
            args = parser.parse_args("a b c".split())
            args.accumulate(args.integers)

    def test_complete_example(self) -> None:
        parser = examples.complete_example()
        args = parser.parse_args("sum 1 2 3".split())

        assert args.func(args.integers) == 6

    def test_compose_clis_using_parents(self) -> None:
        data = {"prog": "daddy", "arguments": [{"name": "something"}]}
        parents = examples.compose_clis_using_parents()
        parser = cli(data, parents=parents)
        args = parser.parse_args(["--foo-parent", "2", "XXX"])

        assert args.something == "XXX"
        assert args.foo_parent == 2

    def test_compose_clis_using_parents_and_arguments(self) -> None:
        data = {"prog": "daddy", "arguments": [{"name": "--something"}]}
        parents = examples.compose_clis_using_parents()
        parser = cli(data, parents=parents)
        args = parser.parse_args(["--something", "XXX"])

        assert args.something == "XXX"

    def test_prefix_chars(self) -> None:
        data = examples.prefix_chars()
        parser = cli(data)
        args = parser.parse_args("+f X ++bar Y".split())

        assert args.foo == "X"
        assert args.bar == "Y"

    def test_name_or_flags(self) -> None:
        data = examples.name_or_flags()
        parser = cli(data)

        args = parser.parse_args(["HELLO"])
        assert args.bar == "HELLO"

        args = parser.parse_args(["BAR", "--foo", "FOO"])
        assert args.bar == "BAR"
        assert args.foo == "FOO"

    def test_name_or_flags_fail(self) -> None:
        data = examples.name_or_flags()
        parser = cli(data)
        with pytest.raises(SystemExit):
            parser.parse_args(["--foo", "FOO"])

    def test_cli_no_args(self) -> None:
        data = {"prog": "daddy", "description": "helloo"}
        parser = cli(data)
        args = parser.parse_args([])

        assert args.__dict__ == {}

    def test_groups(self) -> None:
        data = examples.grouping_arguments()
        parser = cli(data)
        help_result = parser.format_help()

        assert "app" in help_result
        assert "package" in help_result

    def test_not_optional_arg_name_validation_fails(self) -> None:
        data = {"arguments": [{"name": ["f", "foo"]}]}
        with pytest.raises(ValueError):
            cli(data)

    def test_exclusive_group_ok(self) -> None:
        data = {
            "prog": "app",
            "arguments": [
                {
                    "name": "--install",
                    "action": "store_true",
                    "exclusive_group": "ops",
                },
                {
                    "name": "--purge",
                    "action": "store_true",
                    "exclusive_group": "ops",
                },
            ],
        }
        parser = cli(data)
        args = parser.parse_args(["--install"])
        assert args.install is True
        assert args.purge is False

    def test_exclusive_group_fails_when_same_group_called(self) -> None:
        data = {
            "prog": "app",
            "arguments": [
                {
                    "name": "--install",
                    "action": "store_true",
                    "exclusive_group": "opas",
                },
                {
                    "name": "--purge",
                    "action": "store_true",
                    "exclusive_group": "opas",
                },
            ],
        }

        parser = cli(data)
        with pytest.raises(SystemExit):
            parser.parse_args("--install --purge".split())

    def test_exclusive_group_and_group_together_fail(self) -> None:
        """
        Note:

        Exclusive group requires at least one arg before adding groups
        """
        data = {
            "prog": "app",
            "arguments": [
                {
                    "name": "--install",
                    "action": "store_true",
                    "exclusive_group": "ops",
                    "group": "cmd",
                },
                {
                    "name": "--purge",
                    "action": "store_true",
                    "exclusive_group": "ops",
                    "group": "cmd",
                },
                {"name": "--fear", "exclusive_group": "ops"},
            ],
        }

        with pytest.raises(ValueError):
            cli(data)

    def test_subcommand_required(self) -> None:
        data = {
            "prog": "cz",
            "description": (
                "Commitizen is a cli tool to generate conventional commits.\n"
                "For more information about the topic go to "
                "https://conventionalcommits.org/"
            ),
            "formatter_class": argparse.RawDescriptionHelpFormatter,
            "arguments": [
                {"name": "--debug", "action": "store_true", "help": "use debug mode"},
                {
                    "name": ["-n", "--name"],
                    "help": "use the given commitizen (default: cz_conventional_commits)",
                },
            ],
            "subcommands": {
                "title": "commands",
                "required": True,
                "commands": [
                    {
                        "name": ["init"],
                        "help": "init commitizen configuration",
                    }
                ],
            },
        }

        parser = cli(data)
        args = parser.parse_args(["-n", "cz_jira", "init"])
        assert args.debug is False
        assert args.name == "cz_jira"



"""Simple app example"""

from decli import cli

data = {
    "prog": "app",
    "arguments": [
        {"name": "--install", "action": "store_true", "group": "opas"},
        {"name": "--purge", "action": "store_false", "group": "opas"},
    ],
    "subcommands": {
        "title": "main",
        "commands": [
            {
                "name": "commit",
                "arguments": [
                    {
                        "name": "--bocha",
                        "action": "store_true",
                        "group": "opas",
                    }
                ],
            }
        ],
    },
}

parser = cli(data)
args = parser.parse_args()
print(args)



from __future__ import annotations

import logging
from argparse import (
    ArgumentParser,
    _ArgumentGroup,
    _MutuallyExclusiveGroup,
    _SubParsersAction,
)
from collections.abc import Generator, Iterable
from copy import deepcopy
from typing import Any, Callable

config: dict = {}
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def init_config() -> dict[str, Any]:
    return {"prefix_chars": "-"}


def ensure_list(name: str | list[str]) -> list[str]:
    if isinstance(name, str):
        return [name]
    return name


def has_many_and_is_optional(names: list[str]) -> list[str]:
    """The arguments can have aliases only when they are optional.

    If this is not the case, then it raises an error.
    """
    prefix_chars = config["prefix_chars"]
    is_optional = all(name.startswith(tuple(prefix_chars)) for name in names)

    if not is_optional and len(names) > 1:
        raise ValueError(
            f"Only optional arguments (starting with {prefix_chars}) can have aliases"
        )
    return names


def is_exclusive_group_or_not(arg: dict) -> None:
    if "exclusive_group" in arg and "group" in arg:
        raise ValueError("choose group or exclusive_group not both.")


def validate_args(args: Iterable[dict]) -> Generator[dict, Any, None]:
    for arg in args:
        arg["name"] = ensure_list(arg["name"])
        has_many_and_is_optional(arg["name"])
        is_exclusive_group_or_not(arg)
        yield arg


def get_or_create_group(
    parser: ArgumentParser,
    groups: dict,
    title: str | None = None,
    description: str | None = None,
) -> _ArgumentGroup | Any:
    group_parser = groups.get(title)
    if group_parser is None:
        group_parser = parser.add_argument_group(title, description)
        groups.update({title: group_parser})
    return group_parser


def get_or_create_exclusive_group(
    parser: ArgumentParser,
    groups: dict,
    title: str | None = None,
    required: bool = False,
) -> _MutuallyExclusiveGroup | Any:
    group_parser = groups.get(title)
    if group_parser is None:
        group_parser = parser.add_mutually_exclusive_group(required=required)
        groups.update({title: group_parser})

    return group_parser


def add_arguments(parser: ArgumentParser, args: list) -> None:
    groups: dict = {}
    exclusive_groups: dict = {}

    for arg in validate_args(args):
        logger.debug("arg: %s", arg)

        name: list = arg.pop("name")
        group_title: str | None = arg.pop("group", None)
        exclusive_group_title: str | None = arg.pop("exclusive_group", None)

        _parser: Any = parser
        if exclusive_group_title:
            logger.debug("Exclusive group: %s", exclusive_group_title)
            _parser = get_or_create_exclusive_group(
                parser, exclusive_groups, exclusive_group_title
            )
        elif group_title:
            logger.debug("Group: %s", group_title)
            _parser = get_or_create_group(parser, groups, group_title)

        _parser.add_argument(*name, **arg)


def add_subcommand(parser: _SubParsersAction[ArgumentParser], command: dict) -> None:
    args: list = command.pop("arguments", None)
    func: Callable | None = command.pop("func", None)

    names: list = ensure_list(command.pop("name"))
    name: str = names.pop(0)

    if names:
        command.update({"aliases": names})

    command_parser = parser.add_parser(name, **command)

    if func:
        command_parser.set_defaults(func=func)
    if args:
        add_arguments(command_parser, args)


def add_subparser(parser: ArgumentParser, subcommand: dict) -> None:
    commands: list = subcommand.pop("commands")

    # This design is for python 3.6 compatibility
    if "required" in subcommand:
        required = subcommand.pop("required")
        subparser = parser.add_subparsers(**subcommand)
        subparser.required = required
    else:
        subparser = parser.add_subparsers(**subcommand)

    for command in commands:
        add_subcommand(subparser, command)


def add_parser(
    data: dict, parser_class: Callable[..., ArgumentParser], parents: list | None
) -> ArgumentParser:
    if parents is None:
        parents = []

    args: list | None = data.pop("arguments", None)
    subcommands: dict | None = data.pop("subcommands", None)

    parser = parser_class(**data, parents=parents)

    if args:
        logger.debug("Adding arguments...")
        add_arguments(parser, args)

    if subcommands:
        logger.debug("Adding subcommands...")
        add_subparser(parser, subcommands)

    return parser


def cli(
    data: dict,
    parser_class: Callable[..., ArgumentParser] = ArgumentParser,
    parents: list | None = None,
) -> ArgumentParser:
    """Create a cli application.

    This is the entrypoint.
    """
    global config
    config = init_config()
    data = deepcopy(data)

    if data.get("prefix_chars"):
        config.update({"prefix_chars": data.get("prefix_chars")})

    parser = add_parser(data, parser_class, parents)
    return parser
